"""boto3 utilities."""
import logging
import time
import timeit

import boto3

log = logging.getLogger(__name__)


def glue_crawl(crawler: str, *, timeout_minutes: int = 60, retry_seconds: int = 5) -> None:
    """Crawl the specified AWS Glue crawler, waiting until completion."""
    # Ref: https://stackoverflow.com/a/66072347/
    timeout_seconds = timeout_minutes * 60
    client = boto3.client("glue")
    start_time = timeit.default_timer()
    abort_time = start_time + timeout_seconds

    def check_for_timeout() -> None:
        if timeit.default_timer() > abort_time:
            raise TimeoutError(f"Failed to crawl {crawler}. The allocated time of {timeout_minutes:,} minutes has elapsed.")

    def wait_until_ready() -> None:
        state_previous = None
        while True:
            response_get = client.get_crawler(Name=crawler)
            state = response_get["Crawler"]["State"]
            if state != state_previous:
                log.info(f"Crawler {crawler} is {state.lower()}.")
                state_previous = state
            if state == "READY":  # Other known states: RUNNING, STOPPING
                return
            check_for_timeout()
            time.sleep(retry_seconds)

    wait_until_ready()
    response_start = client.start_crawler(Name=crawler)
    assert response_start["ResponseMetadata"]["HTTPStatusCode"] == 200
    log.info(f"Crawling {crawler}.")
    wait_until_ready()
    log.info(f"Crawled {crawler}.")



"""pypolars smart_open utilities."""
import logging
from pathlib import Path
from typing import Any

import pypolars as pl
import smart_open

from wpdatautil.pathlib import path_to_uri
from wpdatautil.timeit import Timer

log = logging.getLogger(__name__)


def read_parquet_path(path: Path, *, df_description: str = "dataframe", **kwargs: Any) -> pl.DataFrame:
    """Return a dataframe read from the given path using `smart_open`.

    :param path: This must be a `pathlib` or `s3path` path.
    :param df_description: Optional description of dataframe.
    :param kwargs: These are forwarded to `pl.read_parquet`.
    """
    uri = path_to_uri(path)
    timer = Timer()
    read_description = f"{df_description} from {uri} using smart-open"
    log.debug(f"Reading {read_description}.")
    try:
        with smart_open.open(uri, "rb") as input_file:
            df = pl.read_parquet(input_file, **kwargs)
    except Exception as exception:  # pylint: disable=broad-except
        log.info(f"Error reading {read_description}: {exception.__class__.__qualname__}: {exception}")
        raise
    log.info(f"Read {read_description} returning {len(df):,} rows in {timer}.")
    return df



"""pathlib utilities."""
import urllib.parse
from pathlib import Path


def ensure_parent_dir(path: Path) -> None:
    """Ensure that the parent directory of the given path exists."""
    path.parent.mkdir(parents=True, exist_ok=True)


def path_to_uri(path: Path) -> str:
    """Return the unquoted URI corresponding to the given path."""
    return urllib.parse.unquote(path.as_uri())  # Note: unquote replaces '%3D' with '=' as is often required.
